---------------------------------------------
-- update saved variable table dynamically --
---------------------------------------------
--Current SavedVar revision
local CURRENT_DATA_VERSION = 1
local function UpdateSavedVariable(data)
	if (not data) then
		--Initial, version 1, table, updated afterwards to match current revision
		data = {
			items={equipment={},bags={},bank={}},
			roles={},
			sets={},
			version=1,
		}
	end
	if data.version == CURRENT_DATA_VERSION then
		return data
	end
	--upgrade per revision
	--ex.:
	--[[if data.version == 1 then
		data.something = {somevalue={}}
		data.version = 2 <-- never forget this or the WoW client will crash in an infinite loop!
	end--]]
	return UpdateSavedVariable(data)
end

--------------------------------
-- load libraries into locals --
--------------------------------
local LibCompress = LibStub:GetLibrary("LibCompress")

---------------------------------
-- define static popup dialogs --
---------------------------------
------------------
-- set creation --
------------------
StaticPopupDialogs["CHECKROLES_NEWSET"] = {
	button1=CREATE,
	button2=BACK,
	OnShow=function(self)
		self.button1:Disable()
	end,
	hasEditBox=true,
	EditBoxOnTextChanged = function(self)
		local t = self:GetText()
		if (CheckRoles_Data.sets[t]) or t == "" then
			self:GetParent().button1:Disable()
		else
			self:GetParent().button1:Enable()
		end
	end,
	EditBoxOnEnterPressed = function(self)
		self:GetParent().button1:Click()
	end,
	OnAccept=function(self)
		local setname = self.editBox:GetText()
		CheckRoles_Data.sets[setname] = {}
		CheckRoles.GearEdit:ApplySet(CheckRoles_Data.sets[setname])
		CheckRoles.GearEdit:UpdateDropdown()
		CheckRoles.GearEdit.dropdown.text:SetText(setname)
	end,
	text="Enter set name:",
	timeout=0,
	whileDead=true,
	hideOnEscape=false,
}

------------------
-- set deletion --
------------------
StaticPopupDialogs["CHECKROLES_DELETESET"]={
	button1=DELETE,
	button2=CANCEL,
	text="Are you sure you want to delete set |cffcccc33%s|r?",
	timeout=0,
	whileDead=true,
	hideOnEscape=false,
	OnAccept=function(self,data)
		CheckRoles_Data.sets[data] = nil
		for _, t in pairs(CheckRoles_Data.roles) do
			for _, s in pairs(t) do
				if s.setName == data then
					s.setName = nil
				end
			end
		end
		CheckRoles:UpdateAllSetDropdowns()
		CheckRoles.GearEdit:UnloadSet()
	end,
	showAlert=true,
}

----------------------------------------
-- copy a table and all its subtables --
----------------------------------------
local function tcopy(intable,outtable,overwrite,written)
	--written is an internal table to track references to already copied tables and keep those intact
	--create outtable if it doesn't exist
	outtable = outtable or {}
	--create written if this is the first call (user call)
	written = written or {}
	--add copied table to written
	written[tostring(intable)] = outtable
	--loop through data
	for key, data in pairs(intable) do
		--if overwrite or not exists, write value
		if (not outtable[key]) or overwrite then
			if type(data) == "table" then
				outtable[key] = written[tostring(data)] or tcopy(data)
			else
				outtable[key] = data
			end
		end
	end
	return outtable
end
CheckRoles = CreateFrame("Frame")
local CheckRoles = CheckRoles

-------------------------------------------
-- hide a frame without invoking scripts --
-------------------------------------------
local function NoScriptHide(f)
	local a = f:GetScript("OnHide")
	f:SetScript("OnHide", nil)
	f:Hide()
	f:SetScript("OnHide", a)
end

--------------------------------
-- loop through texture types --
--------------------------------
local function LoopTextures(t, key)
	t = t or 1
	if type(t) == "number" then
		-- 1 = normal, 2=checkbutton
		return LoopTextures, tostring(t)
	end
	if key == nil then
		return "Normal", "Up"
	end
	if key == "Normal" then
		return "Highlight", "Highlight"
	end
	if key == "Highlight" then
		return "Pushed", "Down"
	end
	if key == "Pushed" then
		return "Disabled", "Disabled"
	end
	if (key == "Disabled") and (t == "2") then
		return "Checked", "Check"
	end
	if (key == "Checked") and (t=="2") then
		return "DisabledChecked", "Check-Disabled"
	end
end

------------------------
-- loop through roles --
------------------------
local function LoopRoles(t, key)
	if (not t) then
		return LoopRoles, true
	end
	if key == nil then
		return "T"
	end
	if key == "T" then
		return "D"
	end
	if key == "D" then
		return "H"
	end
end

------------------------------------------
-- loop through sides (used for insets) --
------------------------------------------
local function LoopSides(t, key)
	if (not t) then
		return LoopSides, true
	end
	if key == nil then
		return "top"
	end
	if key == "top" then
		return "bottom"
	end
	if key == "bottom" then
		return "left"
	end
	if key == "left" then
		return "right"
	end
end

----------------------------------------------
-- loop through inventory (equipment) slots --
----------------------------------------------
local function LoopEquipSlots(t, key)
	if (not t) then
		return LoopEquipSlots, true
	end
	if key == nil then
		return "Head", "Head"
	end
	if key == "Head" then
		return "Neck", "Neck"
	end
	if key == "Neck" then
		return "Shoulder", "Shoulder"
	end
	if key == "Shoulder" then
		return "Cloak", "Chest"
	end
	if key == "Cloak" then
		return "Chest", "Chest"
	end
	if key == "Chest" then
		return "Shirt", "Shirt"
	end
	if key == "Shirt" then
		return "Tabard", "Tabard"
	end
	if key == "Tabard" then
		return "Bracers", "Wrists"
	end
	if key == "Bracers" then
		return "Main Hand", "MainHand"
	end
	if key == "Main Hand" then
		return "Off-hand", "SecondaryHand"
	end
	if key == "Off-hand" then
		if UnitHasRelicSlot("player") then
			return "Relic", "Relic"
		else
			return "Ranged", "Ranged"
		end
	end
	if (key == "Relic" or key == "Ranged") then
		return "Gloves", "Hands"
	end
	if key == "Gloves" then
		return "Belt", "Waist"
	end
	if key == "Belt" then
		return "Legs", "Legs"
	end
	if key == "Legs" then
		return "Boots", "Feet"
	end
	if key == "Boots" then
		return "Top Finger", "Finger"
	end
	if key == "Top Finger" then
		return "Bottom Finger", "Finger"
	end
	if key == "Bottom Finger" then
		return "Top trinket", "Trinket"
	end
	if key == "Top trinket" then
		return "Bottom trinket", "Trinket"
	end
end

-----------------------------------------------
-- loop _all_ equip slots (relic and ranged) --
-----------------------------------------------
local function LoopAllEquipSlots(t, key)
	if (not t) then
		return LoopAllEquipSlots, true
	end
	if key == nil then
		return "Head", "Head"
	end
	if key == "Head" then
		return "Neck", "Neck"
	end
	if key == "Neck" then
		return "Shoulder", "Shoulder"
	end
	if key == "Shoulder" then
		return "Cloak", "Chest"
	end
	if key == "Cloak" then
		return "Chest", "Chest"
	end
	if key == "Chest" then
		return "Shirt", "Shirt"
	end
	if key == "Shirt" then
		return "Tabard", "Tabard"
	end
	if key == "Tabard" then
		return "Bracers", "Wrists"
	end
	if key == "Bracers" then
		return "Main Hand", "MainHand"
	end
	if key == "Main Hand" then
		return "Off-hand", "SecondaryHand"
	end
	if key == "Off-hand" then
		return "Relic", "Relic"
	end
	if key == "Relic" then
		return "Ranged", "Ranged"
	end
	if key == "Ranged" then
		return "Gloves", "Hands"
	end
	if key == "Gloves" then
		return "Belt", "Waist"
	end
	if key == "Belt" then
		return "Legs", "Legs"
	end
	if key == "Legs" then
		return "Boots", "Feet"
	end
	if key == "Boots" then
		return "Top Finger", "Finger"
	end
	if key == "Top Finger" then
		return "Bottom Finger", "Finger"
	end
	if key == "Bottom Finger" then
		return "Top trinket", "Trinket"
	end
	if key == "Top trinket" then
		return "Bottom trinket", "Trinket"
	end
end

---------------------------------------------------
-- get point and relativepoint for a growth type --
---------------------------------------------------
local function FetchPoints(growth)
	growth = growth or "LEFT"
	if growth == "DOWN" then
		return "TOP", "BOTTOM"
	end
	if growth == "UP" then
		return "BOTTOM", "TOP"
	end
	if growth == "LEFT" then
		return "RIGHT", "LEFT"
	end
	if growth == "RIGHT" then
		return "LEFT", "RIGHT"
	end
end

----------------------------------
-- fetch gear data of inv types --
----------------------------------
local function FetchGearList(...)
	local gearTypes = {...}
	local gearReturn = {}
	for _, type in ipairs(gearTypes) do
		if CheckRoles_Data.items.equipment[type] then
			for _, data in ipairs(CheckRoles_Data.items.equipment[type]) do
				if (data.qual >= 2 and data.reqlvl <= UnitLevel("player")) or (type == "INVTYPE_BODY") or (type == "INVTYPE_TABARD") then
					tinsert(gearReturn, {link=data.link, tex=data.tex})
				end
			end
		end
		if CheckRoles_Data.items.bags[type] then
			for _, data in ipairs(CheckRoles_Data.items.bags[type]) do
				if (data.qual >= 2 and data.reqlvl <= UnitLevel("player")) or (type == "INVTYPE_BODY") or (type == "INVTYPE_TABARD") then
					tinsert(gearReturn, {link=data.link, tex=data.tex})
				end
			end
		end
		if CheckRoles_Data.items.bank[type] then
			for _, data in ipairs(CheckRoles_Data.items.bank[type]) do
				if (data.qual >= 2 and data.reqlvl <= UnitLevel("player")) or (type == "INVTYPE_BODY") or (type == "INVTYPE_TABARD") then
					tinsert(gearReturn, {link=data.link, tex=data.tex})
				end
			end
		end
	end
	return gearReturn
end

-----------------------------------------
-- fetch list of invslots from slot ID --
-----------------------------------------
local slotsTypes = {
	Head={"INVTYPE_HEAD"},
	Neck={"INVTYPE_NECK"},
	Shoulder={"INVTYPE_SHOULDER"},
	Cloak={"INVTYPE_CLOAK"},
	Chest={"INVTYPE_CHEST", "INVTYPE_ROBE"},
	Shirt={"INVTYPE_BODY"},
	Tabard={"INVTYPE_TABARD"},
	Bracers={"INVTYPE_WRIST"},
	["Main Hand"]={"INVTYPE_WEAPON", "INVTYPE_2HWEAPON", "INVTYPE_WEAPONMAINHAND"},
	["Off-hand"]={"INVTYPE_WEAPON", "INVTYPE_WEAPONOFFHAND", "INVTYPE_SHIELD", "INVTYPE_HOLDABLE"},
	Gloves={"INVTYPE_HAND"},
	Belt={"INVTYPE_WAIST"},
	Legs={"INVTYPE_LEGS"},
	Boots={"INVTYPE_FEET"},
	["Top Finger"]={"INVTYPE_FINGER"},
	["Bottom Finger"]={"INVTYPE_FINGER"},
	["Top trinket"]={"INVTYPE_TRINKET"},
	["Bottom trinket"]={"INVTYPE_TRINKET"},
	["Relic"] = {"INVTYPE_RELIC"},
	["Ranged"] = {"INVTYPE_RANGED", "INVTYPE_RANGEDRIGHT", "INVTYPE_THROWN"},
}
local function FetchInvTypes(slot)
	return unpack(slotsTypes[slot])
end

----------------------------------------------------
-- get the last value of a table, then set it nil --
----------------------------------------------------
local function CutOffLast(t)
	local a = t[#t]
	t[#t] = nil
	return a
end

------------------------------------------------------------
-- set a value in a table and create all needed subtables --
------------------------------------------------------------
local function CheckSet(t, ...)
	local tables = {...}
	local value = CutOffLast(tables)
	local key = CutOffLast(tables)
	while tables[1] do
		t[tables[1]] = t[tables[1]] or {}
		t = t[tables[1]]
		tremove(tables, 1)
	end
	t[key] = value
end

----------------------------------------------------------------------------------------------
-- get a value from a table while avoiding non-existing subtables and returning nil instead --
----------------------------------------------------------------------------------------------
local function CheckGet(t, ...)
	local tables={...}
	local key = CutOffLast(tables)
	while tables[1] do
		if t[tables[1]] then
			t = t[tables[1]]
		else
			return
		end
		tremove(tables, 1)
	end
	return t[key]
end

-----------------------------------
-- custom frame create functions --
-----------------------------------
-------------------------------------------
-- title bar with close button and label --
-------------------------------------------
local function CreateTitleBar(parent, title, height)
	parent:SetMovable(true)
	height = height or 30
	title = title or parent:GetName()
	local bar = CreateFrame("Frame", nil, parent)
	bar:EnableMouse(true)
	bar:SetScript("OnMouseDown", function() parent:StartMoving() end)
	bar:SetScript("OnMouseUp", function() parent:StopMovingOrSizing() end)
	bar:SetPoint("TOPLEFT")
	bar:SetPoint("TOPRIGHT")
	bar:SetHeight(height)
	bar:SetBackdrop({
		bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
		tile=true,
		tileSize=16,
		edgeSize=16,
		insets={top=4,bottom=4,left=4,right=4},
	})
	bar:SetBackdropBorderColor(0.8,0.8,0.2)
	bar:SetBackdropColor(0.3,0.3,0.3)
	bar.close = CreateFrame("Button", nil, bar)
	bar.close:SetScript("OnClick", function() parent:Hide() end)
	bar.close:RegisterForClicks("LeftButtonUp")
	for textype, state in LoopTextures() do
		bar.close["Set"..textype.."Texture"](bar.close,"Interface\\Buttons\\UI-Panel-MinimizeButton-"..state)
	end
	local addSize= ((height*0.5)*(6/14))
	local totalSize=(height*0.5)+(2*addSize)
	local offSet = addSize/8
	bar.close:SetSize(totalSize,totalSize)
	bar.close:SetHitRectInsets(addSize,addSize,addSize,addSize)
	--bar.close:SetPoint("TOPRIGHT",offSet,offSet)
	bar.close:SetPoint("TOPRIGHT")
	if title then
		bar.text = bar:CreateFontString()
		bar.text:SetFontObject(GameFontHighlight)
		bar.text:SetTextColor(0.9,0.9,0.9)
		bar.text:SetText(title)
	end
	local strh = (height-bar.text:GetStringHeight())/2
	bar.text:SetPoint("TOPLEFT", strh, 0)
	bar.text:SetPoint("BOTTOMLEFT", strh, 0)
	bar.text:SetJustifyH("LEFT")
	return bar
end

------------------
-- dropdown box --
------------------
local function CreateDropdown(p)
	local dropdown = CreateFrame("Frame", nil,p)
	dropdown:SetBackdrop({
		bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
		edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
		tile=true,
		tileSize=16,
		edgeSize=8,
		insets={top=4,bottom=4,right=4,left=4},
	})
	dropdown:SetScript("OnSizeChanged", function(self, width, height)
		local drp = self:GetBackdrop()
		drp.edgeSize=height*(12/20)
		drp.tileSize=height*(16/20)
		local inset = height*(4/20)
		for side in LoopSides() do
			drp.insets[side] = inset
		end
		self:SetBackdrop(drp)
		self.button:SetSize(height*(18/20),height*(18/20))
		self.button:ClearAllPoints()
		self.button:SetPoint("RIGHT", -0.5*drp.edgeSize/4, 0)
		if height < 25 then
			self.text:SetFontObject(GameFontHighlightSmall)
		else
			self.text:SetFontObject(GameFontHighlight)
		end
	end)
	dropdown.button = CreateFrame("Button", nil, dropdown)
	for textype, state in LoopTextures(1) do
		if textype == "Highlight" then
			dropdown.button:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
		else
			dropdown.button["Set"..textype.."Texture"](dropdown.button, "Interface\\ChatFrame\\UI-ChatIcon-ScrollDown-"..state)
		end
	end
	dropdown.GetValue = function() end
	dropdown.button:Disable()
	dropdown.listframe = CreateFrame("Frame", nil, dropdown)
	dropdown.listframe:Hide()
	dropdown.listframe:SetPoint("TOPLEFT", dropdown, "BOTTOMLEFT")
	dropdown.listframe:SetPoint("TOPRIGHT", dropdown, "BOTTOMRIGHT")
	dropdown.listframe:SetBackdrop({
		bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
		edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
		tile=true,
		tileSize=16,
		edgeSize=8,
		insets={top=4,bottom=4,right=4,left=4},
	})
	dropdown.listframe.entries = setmetatable({}, {__index=function(self, key)
		local f = CreateFrame("Button", nil, dropdown.listframe)
		if key > 1 then
			f:SetPoint("TOPLEFT", self[key-1], "BOTTOMLEFT", 0, -2)
			f:SetPoint("TOPRIGHT", self[key-1], "BOTTOMRIGHT", 0, -2)
		else
			f:SetPoint("TOPLEFT", 2, -2)
			f:SetPoint("TOPRIGHT", -2, -2)
		end
		f:SetHeight(20)
		f:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
		f:RegisterForClicks("AnyUp")
		f:SetScript("OnClick", function(self)
			if (not self.callback) or (not self.key) or (not self.text) then return end
			self.callback()
			dropdown.GetValue = function() return self.key end
			dropdown.text:SetText(self.text)
			dropdown.listframe:Hide()
		end)
		function f:SetData(key, callback, text)
			self.fontstring:SetText(text)
			self.text = text
			self.key = key
			self.callback = callback
		end
		f.fontstring = f:CreateFontString()
		local stxt = f.fontstring.SetText
		function f.fontstring:SetText(text)
			stxt(self,text)
			local a = self:GetStringHeight()
			self:ClearAllPoints()
			self:SetPoint("LEFT", (20-a)/2, 0)
		end
		f.fontstring:SetFontObject(GameFontHighlightSmallLeft)
		self[key] = f
		return f
	end})
	function dropdown:SetList(list)
		dropdown.button:Enable()
		--[[
		list format is as follows:
		{
			[numeric id]={text=text, callback=callback, key=key},
		}
		--]]
		for _, frame in pairs(self.listframe.entries) do
			frame:Hide()
		end
		for numid, data in ipairs(list) do
			local btn = self.listframe.entries[numid]
			btn:Show()
			btn:SetData(data.key, data.callback, data.text)
		end
		self.listframe:SetHeight(2+(#list*22))
		self.text:SetText("")
	end
	dropdown.button:SetScript("OnClick", function()
		dropdown.listframe[dropdown.listframe:IsShown() and "Hide" or "Show"](dropdown.listframe)
	end)
	dropdown.text = dropdown:CreateFontString()
	dropdown.text:SetFontObject(GameFontHighlight)
	do
		local st = dropdown.text.SetText
		function dropdown.text:SetText(text)
			st(self,text)
			self:ClearAllPoints()
			self:SetPoint("LEFT", (self:GetParent():GetHeight()-self:GetStringHeight())/2, 0)
			function dropdown.GetValue() return text end
		end
	end
	return dropdown
end

------------------------------------------------------
-- create all gui frames and register slash command --
------------------------------------------------------
function CheckRoles:InitGUI()
	self.classRoles = CheckRoles_Roles[(select(2, UnitClass("player")))]
	do
		self.RoleSelect = CreateFrame("Frame", nil, UIParent)
		local roleselect = self.RoleSelect
		roleselect:SetScript("OnHide", function(self) if self.ShouldShowActionSelectOnHide then self.ShouldShowActionSelectOnHide = false CheckRoles.ActionSelect:Show() end end)
		roleselect:Hide()
		roleselect:SetPoint("CENTER")
		roleselect.title = CreateTitleBar(roleselect, "Role select")
		roleselect.roleframes = setmetatable({}, {__index=function(t, key)
			local f = CreateFrame("Frame", nil, roleselect)
			if key > 1 then
				f:SetPoint("TOPLEFT", t[key-1], "TOPRIGHT")
				f:SetPoint("BOTTOMLEFT", t[key-1], "BOTTOMRIGHT")
			else
				f:SetPoint("TOPLEFT", roleselect.title, "BOTTOMLEFT")
				f:SetPoint("BOTTOMLEFT")
			end
			f:SetWidth(180)
			roleselect:SetWidth(180*key)
			f:SetBackdrop({
				bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
				edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
				tile=true,
				tileSize=16,
				edgeSize=16,
				insets={top=4,bottom=4,right=4,left=4},
			})
			f:SetBackdropColor(0.5,0.5,0.5)
			f:SetBackdropBorderColor(0.8,0.8,0.2)
			f.icon = f:CreateTexture()
			f.icon:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-ROLES")
			f.icon:SetSize(40,40)
			f.icon:SetPoint("TOP", 0, -10)
			f.icontext = f:CreateFontString()
			f.icontext:SetFont("Fonts\\FRIZQT__.TTF", 12, "OUTLINE")
			f.icontext:SetTextColor(0.8,0.8,0.2)
			f.icontext:SetPoint("CENTER", f.icon, "CENTER")
			local sd = f.icon.SetDesaturated
			function f.icon:SetDesaturated(desaturate)
				if not sd(self,desaturate) then
					local v = desaturate and 0.5 or 1
					self:SetVertexColor(v,v,v)
				end
			end
			f.icon:SetDesaturated(true)
			function f:UpdateDesaturation()
				for _, fr in pairs(f.specs) do
					if fr.checkbox:GetChecked() then
						self.icon:SetDesaturated(false)
						return
					end
				end
				self.icon:SetDesaturated(true)
			end
			f.specs = setmetatable({}, {__index=function(t2, key2)
				local f2 = CreateFrame("Frame", nil, f)
				if key2 > 1 then
					f2:SetPoint("TOPLEFT", t2[key2-1], "BOTTOMLEFT", 0, -3)
					f2:SetPoint("TOPRIGHT", t2[key2-1], "BOTTOMRIGHT", 0, -3)
				else
					f2:SetPoint("TOPLEFT", 5, -53)
					f2:SetPoint("TOPRIGHT", -5, -53)
				end
				f2:SetHeight(16)
				local newSize = 88+(key2*19)
				if newSize > roleselect:GetHeight() then
					roleselect:SetHeight(newSize)
				end
				f2.checkbox = CreateFrame("CheckButton", nil, f2)
				f2.checkbox:SetPoint("LEFT")
				f2.checkbox:SetSize(16,16)
				for textype, texstring in LoopTextures(2) do
					f2.checkbox["Set"..textype.."Texture"](f2.checkbox, "Interface\\Buttons\\UI-CheckBox-"..texstring)
				end
				f2.checkbox:SetScript("PostClick", function(self)
					CheckSet(CheckRoles_Data.roles, f.role, f2.spec, "IsEnabled", self:GetChecked() or false)
					f:UpdateDesaturation()
				end)
				f2.roleicon = f2:CreateTexture()
				f2.roleicon:SetPoint("LEFT", f2.checkbox, "RIGHT", 2, 0)
				f2.roleicon:SetSize(16,16)
				f2.roletext=f2:CreateFontString()
				f2.roletext:SetFont("Fonts\\FRIZQT__.TTF", 8)
				f2.roletext:SetTextColor(0.8,0.8,0.2)
				f2.roletext:SetPoint("LEFT", f2.roleicon, "RIGHT", 2, 0)
				f2.dropdown = CreateDropdown(f2)
				f2.dropdown:SetPoint("TOPLEFT", f2.roletext, "TOPRIGHT", 2, 0)
				f2.dropdown:SetPoint("BOTTOMRIGHT")
				function f2:SetSpec(specID)
					self.roleicon:SetTexture(CheckRoles.speclocal[specID].icon)
					self.roletext:SetText(CheckRoles.speclocal[specID].name)
					self.spec = specID
					self.checkbox:SetChecked(CheckGet(CheckRoles_Data.roles, f.role, specID, "IsEnabled") or false)
					f:UpdateDesaturation()
				end
				function f2:UpdateDropdown()
					local t = {}
					for setname, setdata in pairs(CheckRoles_Data.sets) do
						tinsert(t, {text=setname, key=setname, callback=function() CheckSet(CheckRoles_Data.roles, f.role, f2.spec, "setName", setname)  end})
					end
					self.dropdown:SetList(t)
					self.dropdown.text:SetText(CheckGet(CheckRoles_Data.roles, f.role, f2.spec, "setName") or "")
				end
				f2:SetScript("OnShow", f2.UpdateDropdown)
				t2[key2]=f2
				return f2
			end})
			function f:SetRole(role)
				f.role = role
				f.icon:SetTexCoord(GetTexCoordsForRole(role=="T" and "TANK" or role=="D" and "DAMAGER" or role=="H" and "HEALER"))
				f.icontext:SetText(role=="T" and "Tank" or role=="D" and "DPS" or role=="H" and "Healer")
				local cnt = 1
				for specNum, isRole in ipairs(CheckRoles.classRoles[role]) do
					if isRole then
						local btn = self.specs[cnt]
						btn:SetSpec(specNum)
						cnt=cnt+1
					end
				end
			end
			t[key] = f
			return f
		end})
		local cnt = 1
		for role in LoopRoles() do
			if self.classRoles[role] then
				local roleframe = roleselect.roleframes[cnt]
				roleframe:SetRole(role)
				cnt = cnt+1
			end
		end
	end
	do
		self.GearEdit = CreateFrame("Frame", nil, UIParent)
		local gearedit = self.GearEdit
		gearedit:SetScript("OnHide", function(self) if self.ShouldShowActionSelectOnHide then self.ShouldShowActionSelectOnHide = false CheckRoles.ActionSelect:Show() end end)
		gearedit:Hide()
		gearedit:SetPoint("CENTER")
		gearedit.title = CreateTitleBar(gearedit, "Gear set editor")
		gearedit:SetWidth(300)
		gearedit:SetHeight(400)
		gearedit:SetBackdrop({
			bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
			edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
			tile=true,
			tileSize=16,
			edgeSize=16,
			insets={top=4,bottom=4,left=4,right=4},
		})
		gearedit:SetBackdropColor(0.5,0.5,0.5)
		gearedit.dropdown = CreateDropdown(gearedit)
		gearedit.dropdown:SetPoint("TOPLEFT", gearedit.title, "BOTTOMLEFT", 5, 0)
		gearedit.dropdown:SetPoint("TOPRIGHT", gearedit.title, "BOTTOMRIGHT", -38, 0)
		gearedit.dropdown:SetHeight(30)
		function gearedit:UpdateDropdown()
			local t = {}
			for setname, setdata in pairs(CheckRoles_Data.sets) do
				tinsert(t, {text=setname, key=setname, callback=function() self.curset = setname self:ApplySet(setdata) end})
			end
			tinsert(t, {text="Create new set...", key="newset", callback=function() StaticPopup_Show("CHECKROLES_NEWSET") end})
			self.dropdown:SetList(t)
			if self.curset then self.dropdown.text:SetText(self.curset) end
		end
		gearedit.deleteset = CreateFrame("Button", nil, gearedit)
		gearedit.deleteset:SetPoint("LEFT", gearedit.dropdown, "RIGHT", 2, 0)
		gearedit.deleteset:SetSize(30,30)
		for textype, state in LoopTextures() do
			gearedit.deleteset["Set"..textype.."Texture"](gearedit.deleteset,"Interface\\BUTTONS\\UI-GroupLoot-Pass-"..state)
		end
		gearedit.deleteset:SetDisabledTexture("Interface\\BUTTONS\\UI-GroupLoot-Pass-Up")
		if not gearedit.deleteset:GetDisabledTexture():SetDesaturated(true) then
			gearedit.deleteset:GetDisabledTexture():SetVertexColor(0.5,0.5,0.5)
		end
		gearedit.deleteset:Disable()
		gearedit.deleteset:RegisterForClicks("LeftButtonUp")
		gearedit.deleteset:SetScript("OnClick", function(self)
			StaticPopup_Show("CHECKROLES_DELETESET", gearedit.dropdown:GetValue()).data = gearedit.dropdown:GetValue()
		end)
		gearedit.dressup = CreateFrame("DressUpModel", nil, gearedit)
		gearedit.dressup:SetPoint("TOP", gearedit.dropdown, "BOTTOM", 16, 0)
		gearedit.dressup:SetSize(223,297)
		gearedit.dressup:SetUnit("player")
		gearedit.equipbuttons = {}
		function gearedit:UpdateGear()
			self.dressup:RefreshUnit()
			self.dressup:Undress()
			for _, btn in pairs(self.equipbuttons) do
				if btn.link then
					self.dressup:TryOn(btn.link)
				end
			end
		end
		gearedit:SetScript("OnShow", function(self) self:UpdateGear() self:UpdateDropdown() end)
		function gearedit:DisableGearChanges()
			for _, b in pairs(self.equipbuttons) do
				b:Disable()
			end
		end
		function gearedit:EnableGearChanges()
			for _, b in pairs(self.equipbuttons) do
				b:Enable()
			end
		end
		for slotID, textureData in LoopEquipSlots() do
			local equipbtn = CreateFrame("Button", nil, gearedit)
			do
				local snt = equipbtn.SetNormalTexture
				function equipbtn:SetNormalTexture(...)
					self:SetDisabledTexture(...)
					snt(self, ...)
				end
			end
			equipbtn:SetSize(35,35)
			equipbtn:SetNormalTexture("Interface\\PaperDoll\\UI-PaperDoll-Slot-"..textureData)
			equipbtn:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square")
			equipbtn.slot = slotID
			equipbtn:SetScript("OnClick", function(self)
				if gearedit.curPopout == self.popoutbg then
					self.popoutbg:Hide()
					gearedit.curPopout = nil
					return
				end
				if gearedit.curPopout then
					gearedit.curPopout:Hide()
				end
				gearedit.curPopout = self.popoutbg
				for _, f in ipairs(self.popout) do f:Hide() end
				local list = FetchGearList(FetchInvTypes(self.slot))
				local count = 1
				while list[1] do
					local btn = self.popout[count]
					btn:Show()
					btn:SetData(list[1].link, list[1].tex)
					tremove(list, 1)
					count=count+1
				end
				local btn = self.popout[count]
				btn:Show()
				btn:NoData()
				self.popoutbg:SetWidth(50*count)
				self.popoutbg:Show()
			end)
			equipbtn:SetScript("OnEnter", function(self)
				local gt = GameTooltip
				gt:SetOwner(self, "ANCHOR_RIGHT")
				gt:ClearLines()
				if self.link then
					gt:SetHyperlink(self.link)
				else
					gt:AddLine(self.slot, 0.8,0.8,0.2)
				end
				gt:Show()
			end)
			equipbtn:SetScript("OnLeave", function()
				GameTooltip:Hide()
			end)
			equipbtn.popoutbg = CreateFrame("Frame", nil, equipbtn)
			local point, relativeTo = FetchPoints(equipbtn.growth)
			equipbtn.popoutbg:SetPoint(point, equipbtn, relativeTo)
			equipbtn.popoutbg:SetHeight(50)
			equipbtn.popout = setmetatable({}, {__index=function(tab,key)
				local btn = CreateFrame("Button", nil, equipbtn.popoutbg)
				btn:SetSize(35,35)
				local point, relativeTo = FetchPoints(equipbtn.growth)
				if key > 1 then
					btn:SetPoint(point, tab[key-1], relativeTo)
				else
					btn:SetPoint(point)
				end
				function btn:SetData(link, texture)
					self.link = link
					self.texture = texture
					self:SetNormalTexture(texture)
				end
				function btn:NoData()
					self.link = nil
					self.texture = "Interface\\PaperDoll\\UI-PaperDoll-Slot-"..textureData
					self:SetNormalTexture("Interface\\PaperDoll\\UI-PaperDoll-Slot-"..textureData)
				end
				btn:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square")
				btn:SetScript("OnEnter", function(self)
					local gt = GameTooltip
					gt:SetOwner(self, "ANCHOR_RIGHT")
					gt:ClearLines()
					if self.link then
						gt:SetHyperlink(self.link)
					else
						gt:AddLine("Empty", 0.8,0.8,0.2)
					end
					gt:Show()
				end)
				btn:SetScript("OnLeave", function()
					GameTooltip:Hide()
				end)
				btn:SetScript("OnClick", function(self)
					equipbtn:SetNormalTexture(self.texture)
					equipbtn.link = self.link
					equipbtn.popoutbg:Hide()
					gearedit:UpdateGear()
					CheckSet(CheckRoles_Data.sets, gearedit.dropdown:GetValue(), self.slot, {link=self.link, tex=self.texture})
				end)
				btn.slot = equipbtn.slot
				tab[key] = btn
				return btn
			end})
			function equipbtn:SetGrowth(growthType)
				local point, relativeTo = FetchPoints(growthType)
				for key, frame in ipairs(self.popout) do
					frame:ClearAllPoints()
					if key > 1 then
						frame:SetPoint(point, self.popout[key-1], relativeTo)
					else
						frame:SetPoint(point)
					end
				end
				self.popoutbg:ClearAllPoints()
				self.popoutbg:SetPoint(point, self, relativeTo)
				self.growth = growthType
			end
			gearedit.equipbuttons[slotID] = equipbtn
		end
		gearedit:DisableGearChanges()
		local btns = gearedit.equipbuttons
		btns.Head:SetPoint("TOPRIGHT", gearedit.dressup, "TOPLEFT")
		btns.Head:SetGrowth("LEFT")
		btns.Neck:SetPoint("TOP", btns.Head, "BOTTOM")
		btns.Neck:SetGrowth("LEFT")
		btns.Shoulder:SetPoint("TOP", btns.Neck, "BOTTOM")
		btns.Shoulder:SetGrowth("LEFT")
		btns.Cloak:SetPoint("TOP", btns.Shoulder, "BOTTOM")
		btns.Cloak:SetGrowth("LEFT")
		btns.Chest:SetPoint("TOP", btns.Cloak, "BOTTOM")
		btns.Chest:SetGrowth("LEFT")
		btns.Shirt:SetPoint("TOP", btns.Chest, "BOTTOM")
		btns.Shirt:SetGrowth("LEFT")
		btns.Tabard:SetPoint("TOP", btns.Shirt, "BOTTOM")
		btns.Tabard:SetGrowth("LEFT")
		btns.Bracers:SetPoint("TOP", btns.Tabard, "BOTTOM")
		btns.Bracers:SetGrowth("LEFT")
		btns["Main Hand"]:SetPoint("TOPLEFT", gearedit.dressup, "BOTTOMLEFT")
		btns["Main Hand"]:SetGrowth("UP")
		btns["Off-hand"]:SetPoint("TOP", gearedit.dressup, "BOTTOM")
		btns["Off-hand"]:SetGrowth("UP")
		do
			local a=btns[UnitHasRelicSlot("player") and "Relic" or "Ranged"]
			a:SetPoint("TOPRIGHT", gearedit.dressup, "BOTTOMRIGHT")
			a:SetGrowth("UP")
		end
		btns.Gloves:SetPoint("TOPLEFT", gearedit.dressup, "TOPRIGHT")
		btns.Gloves:SetGrowth("RIGHT")
		btns.Belt:SetPoint("TOP", btns.Gloves, "BOTTOM")
		btns.Belt:SetGrowth("RIGHT")
		btns.Legs:SetPoint("TOP", btns.Belt, "BOTTOM")
		btns.Legs:SetGrowth("RIGHT")
		btns.Boots:SetPoint("TOP", btns.Legs, "BOTTOM")
		btns.Boots:SetGrowth("RIGHT")
		btns["Top Finger"]:SetPoint("TOP", btns.Boots, "BOTTOM")
		btns["Top Finger"]:SetGrowth("RIGHT")
		btns["Bottom Finger"]:SetPoint("TOP", btns["Top Finger"], "BOTTOM")
		btns["Bottom Finger"]:SetGrowth("RIGHT")
		btns["Top trinket"]:SetPoint("TOP", btns["Bottom Finger"], "BOTTOM")
		btns["Top trinket"]:SetGrowth("RIGHT")
		btns["Bottom trinket"]:SetPoint("TOP", btns["Top trinket"], "BOTTOM")
		btns["Bottom trinket"]:SetGrowth("RIGHT")
		function gearedit:ApplySet(data)
			for slot, texdata in LoopEquipSlots() do
				if data[slot] then
					btns[slot]:SetNormalTexture(data[slot].tex)
					btns[slot].link = data[slot].link
				else
					btns[slot]:SetNormalTexture("Interface\\PaperDoll\\UI-PaperDoll-Slot-"..texdata)
					btns[slot].link = nil
				end
			end
			self:EnableGearChanges()
			self.deleteset:Enable()
			self:UpdateGear()
		end
		function gearedit:UnloadSet()
			for slot, texdata in LoopEquipSlots() do
				btns[slot]:SetNormalTexture("Interface\\PaperDoll\\UI-PaperDoll-Slot-"..texdata)
				btns[slot].link = nil
			end
			self:DisableGearChanges()
			self.deleteset:Disable()
			self:UpdateGear()
			self.dropdown.text:SetText("")
		end
	end
	do
		self.DataView = CreateFrame("Frame", nil, UIParent)
		local dataview = self.DataView
		dataview:SetScript("OnHide", function(self) if self.ShouldShowActionSelectOnHide then self.ShouldShowActionSelectOnHide = false CheckRoles.ActionSelect:Show() end end)
		dataview:Hide()
		dataview:SetPoint("CENTER")
		dataview.title = CreateTitleBar(dataview, "Data manager")
		dataview:SetWidth(600)
		dataview:SetHeight(300)
		dataview:SetBackdrop({
			bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
			edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
			tile=true,
			tileSize=16,
			edgeSize=16,
			insets={top=4,bottom=4,left=4,right=4},
		})
		dataview:SetBackdropColor(0.5,0.5,0.5)
		dataview.players = CreateFrame("Frame", nil, dataview)
		dataview.players:SetHeight(250)
		dataview.players:SetWidth(150)
		dataview.players:SetBackdrop({
			bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
			edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
			tile=true,
			tileSize=16,
			edgeSize=16,
			insets={top=4,bottom=4,right=4,left=4},
		})
		dataview.players:SetBackdropColor(0.46,0.46,0.46)
		dataview.players:SetBackdropBorderColor(0.48,0.48,0.48)
		dataview.players:SetPoint("LEFT", 15, -15)
		dataview.players.scrollframe = CreateFrame("ScrollFrame", nil, dataview.players)
		dataview.players.scrollframe:SetPoint("TOPLEFT", 5, -5)
		dataview.players.scrollframe:SetPoint("BOTTOMRIGHT", -5, 5)
		dataview.players.scroll = CreateFrame("Frame", nil, dataview.players)
		dataview.players.scroll:EnableMouseWheel(true)
		dataview.players.scroll:SetScript("OnMouseWheel", function(self, delta)
			if delta == 1 then
				dataview.players.scrollframe:SetVerticalScroll(math.max(dataview.players.scrollframe:GetVerticalScroll()-5,0))
			else
				dataview.players.scrollframe:SetVerticalScroll(math.min(dataview.players.scrollframe:GetVerticalScroll()+5,dataview.players.scrollframe:GetVerticalScrollRange()))
			end
		end)
		dataview.players.scroll:SetWidth(140)
		dataview.players.scrollframe:SetScrollChild(dataview.players.scroll)
		dataview.players.entries = setmetatable({}, {__index=function(tab, key)
			local f = CreateFrame("Button", nil, dataview.players.scroll)
			if key > 1 then
				f:SetPoint("TOPLEFT", tab[key-1], "BOTTOMLEFT")
				f:SetPoint("TOPRIGHT", tab[key-1], "BOTTOMRIGHT")
			else
				f:SetPoint("TOPLEFT")
				f:SetPoint("TOPRIGHT")
			end
			f:SetHeight(20)
			f:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
			f.tank = f:CreateTexture()
			f.tank:SetSize(20,20)
			f.tank:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-ROLES")
			f.tank:SetTexCoord(GetTexCoordsForRole("TANK"))
			f.dps = f:CreateTexture()
			f.dps:SetSize(20,20)
			f.dps:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-ROLES")
			f.dps:SetTexCoord(GetTexCoordsForRole("DAMAGER"))
			f.heal = f:CreateTexture()
			f.heal:SetSize(20,20)
			f.heal:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-ROLES")
			f.heal:SetTexCoord(GetTexCoordsForRole("HEALER"))
			f.tank:SetPoint("LEFT")
			f.dps:SetPoint("LEFT", f.tank, "RIGHT")
			f.heal:SetPoint("LEFT", f.dps, "RIGHT")
			do
				local sd = f.tank.SetDesaturated
				function f.tank:SetDesaturated(desaturate)
					if not sd(self,desaturate) then
						local v = desaturate and 0.5 or 1
						self:SetVertexColor(v,v,v)
					end
				end
			end
			do
				local sd = f.dps.SetDesaturated
				function f.dpsSetDesaturated(desaturate)
					if not sd(self,desaturate) then
						local v = desaturate and 0.5 or 1
						self:SetVertexColor(v,v,v)
					end
				end
			end
			do
				local sd = f.heal.SetDesaturated
				function f.heal:SetDesaturated(desaturate)
					if not sd(self,desaturate) then
						local v = desaturate and 0.5 or 1
						self:SetVertexColor(v,v,v)
					end
				end
			end
			f.name = f:CreateFontString()
			f.name:SetFont("Fonts\\FRIZQT__.TTF", 10)
			f.name:SetTextColor(1,1,1)
			f.name:SetPoint("LEFT", f.heal, "RIGHT", 2, 0)
			f.name:SetJustifyH("LEFT")
			f.name:SetPoint("RIGHT")
			function f:SetData(name, class, data)
				--[[ data:
				{
					[role]={
						[spec]={
							texture=texture,
							data=data,
						},
					},
				}
				--]]
				self.name:SetText(name)
				self.named = name
				local colors = RAID_CLASS_COLORS[class]
				self.name:SetTextColor(colors.r,colors.g,colors.b)
				self.tank:SetDesaturated(not data.T)
				self.dps:SetDesaturated(not data.D)
				self.heal:SetDesaturated(not data.H)
				self.data = data
				self.class = class
			end
			f:SetScript("OnClick", function(self)
				if dataview.players.curHilight then dataview.players.curHilight:UnlockHighlight() end
				dataview.players.curHilight = self
				self:LockHighlight()
				for _, b in ipairs(dataview.specs.entries) do
					b:Hide()
				end
				local cnt = 1
				for role in LoopRoles() do
					if self.data[role] then
						for spec=1, GetNumTalentTabs() do
							if self.data[role][spec] then
								local btn = dataview.specs.entries[cnt]
								btn:Show()
								btn:SetData(self.named, self.data[role][spec].name, self.data[role][spec].texture, self.data[role][spec].data, self.class, role)
								cnt = cnt+1
							end
						end
					end
				end
				dataview.specs.scroll:SetHeight((cnt-1)*20)
				dataview.specs.scrollframe:UpdateScrollChildRect()
				dataview.specs.scrollframe:SetVerticalScroll(0)
			end)
			tab[key] = f
			return f
		end})
		dataview.specs = CreateFrame("Frame", nil, dataview)
		dataview.specs:SetHeight(250)
		dataview.specs:SetWidth(300)
		dataview.specs:SetBackdrop({
			bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
			edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
			tile=true,
			tileSize=16,
			edgeSize=16,
			insets={top=4,bottom=4,right=4,left=4},
		})
		dataview.specs:SetBackdropColor(0.46,0.46,0.46)
		dataview.specs:SetBackdropBorderColor(0.48,0.48,0.48)
		dataview.specs:SetPoint("LEFT", dataview.players, "RIGHT", 15, 0)
		dataview.specs.scrollframe = CreateFrame("ScrollFrame", nil, dataview.specs)
		dataview.specs.scrollframe:SetPoint("TOPLEFT", 5, -5)
		dataview.specs.scrollframe:SetPoint("BOTTOMRIGHT", -5, 5)
		dataview.specs.scroll = CreateFrame("Frame", nil, dataview.specs)
		dataview.specs.scroll:EnableMouseWheel(true)
		dataview.specs.scroll:SetScript("OnMouseWheel", function(self, delta)
			if delta == 1 then
				dataview.specs.scrollframe:SetVerticalScroll(math.max(dataview.specs.scrollframe:GetVerticalScroll()-5,0))
			else
				dataview.specs.scrollframe:SetVerticalScroll(math.min(dataview.specs.scrollframe:GetVerticalScroll()+5,dataview.specs.scrollframe:GetVerticalScrollRange()))
			end
		end)
		dataview.specs.scroll:SetWidth(290)
		dataview.specs.scrollframe:SetScrollChild(dataview.specs.scroll)
		dataview.specs.entries = setmetatable({}, {__index=function(tab, key)
			local f = CreateFrame("Button", nil, dataview.specs.scroll)
			if key > 1 then
				f:SetPoint("TOPLEFT", tab[key-1], "BOTTOMLEFT")
				f:SetPoint("TOPRIGHT", tab[key-1], "BOTTOMRIGHT")
			else
				f:SetPoint("TOPLEFT")
				f:SetPoint("TOPRIGHT")
			end
			f:SetHeight(20)
			f:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
			f.role = f:CreateTexture()
			f.role:SetSize(20,20)
			f.role:SetTexture("Interface\\LFGFrame\\UI-LFG-ICON-ROLES")
			f.spec = f:CreateTexture()
			f.spec:SetSize(20,20)
			f.role:SetPoint("LEFT")
			f.spec:SetPoint("LEFT", f.role, "RIGHT", 2, 0)
			f.specname = f:CreateFontString()
			f.specname:SetFont("Fonts\\FRIZQT__.TTF", 10)
			f.specname:SetTextColor(1,1,1)
			f.specname:SetPoint("LEFT", f.spec, "RIGHT", 2, 0)
			f.specname:SetPoint("RIGHT")
			f.specname:SetJustifyH("LEFT")
			function f:SetData(name, specname, spectex, data, class, role)
				self.specname:SetText(specname)
				self.spec:SetTexture(spectex)
				self.role:SetTexCoord(GetTexCoordsForRole(role=="T" and "TANK" or role=="D" and "DAMAGER" or role=="H" and "HEALER"))
				self.sname = specname
				self.name = name
				self.data = data
				self.class = class
			end
			f:SetScript("OnClick", function(self)
				CheckRoles.GearView:Show()
				CheckRoles.GearView.ShouldShowDataViewOnClose = true
				NoScriptHide(CheckRoles.DataView)
				CheckRoles.GearView.dressup:SetUnit(self.name)
				CheckRoles.GearView:ApplySet(self.data, self.name.." - "..self.sname, (self.class == "PALADIN" or self.class == "SHAMAN" or self.class == "DRUID") and true or false)
				if dataview.specs.curHilight then dataview.specs.curHilight:UnlockHighlight() end
				dataview.specs.curHilight = self
				self:LockHighlight()
			end)
			tab[key] = f
			return f
		end})
		dataview.requestraid = CreateFrame("Button", nil, dataview)
		for textype, state in LoopTextures() do
			dataview.requestraid["Set"..textype.."Texture"](dataview.requestraid, "Interface\\Buttons\\UI-Panel-Button-"..state)
			dataview.requestraid["Get"..textype.."Texture"](dataview.requestraid):SetTexCoord(0,0,0,0.6875,0.625,0,0.625,0.6875)
		end
		dataview.requestraid:SetNormalFontObject(GameFontHighlight)
		dataview.requestraid:SetText("RAID")
		dataview.requestraid:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
			GameTooltip:ClearLines()
			GameTooltip:AddLine("Click to request data from raid")
			GameTooltip:AddLine("Shift-click to add data from raid to current data")
			GameTooltip:Show()
		end)
		dataview.requestraid:SetScript("OnLeave", function()
			GameTooltip:Hide()
		end)
		dataview.requestraid:SetScript("OnClick", function()
			if not (IsLeftShiftKeyDown() or IsRightShiftKeyDown()) then
				table.wipe(CheckRoles.SetDataList)
			end
			CheckRoles:SendCommMessage("CheckRolesRequest", "request", "RAID", "ALERT")
		end)
		dataview.requestraid:RegisterForClicks("LeftButtonUp")
		dataview.requestraid:SetPoint("TOPLEFT", dataview.specs, "TOPRIGHT", 15, 0)
		dataview.requestraid:SetPoint("TOPRIGHT", -15, -15)
		dataview.requestraid:SetHeight(20)
		dataview.requestraid:SetScript("OnEvent", function(self)
			self.update.recheck = GetTime()+0.2
			self.update:Show()
		end)
		dataview.requestraid.update = CreateFrame("Frame", dataview.requestraid)
		dataview.requestraid.update:SetScript("OnUpdate", function()
			if GetTime() >= (self.recheck or 0) then
				self:Hide()
				if GetRealNumRaidMembers() > 0 then
					dataview.requestraid:Enable()
				else
					dataview.requestraid:Disable()
				end
			end
		end)
		dataview.requestraid:RegisterEvent("RAID_ROSTER_UPDATE")
		dataview.requestguild = CreateFrame("Button", nil, dataview)
		for textype, state in LoopTextures() do
			dataview.requestguild["Set"..textype.."Texture"](dataview.requestguild, "Interface\\Buttons\\UI-Panel-Button-"..state)
			dataview.requestguild["Get"..textype.."Texture"](dataview.requestguild):SetTexCoord(0,0,0,0.6875,0.625,0,0.625,0.6875)
		end
		dataview.requestguild:SetNormalFontObject(GameFontHighlight)
		dataview.requestguild:SetText("GUILD")
		dataview.requestguild:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
			GameTooltip:ClearLines()
			GameTooltip:AddLine("Click to request data from guild")
			GameTooltip:AddLine("Shift-click to add data from guild to current data")
			GameTooltip:Show()
		end)
		dataview.requestguild:SetScript("OnLeave", function()
			GameTooltip:Hide()
		end)
		dataview.requestguild:SetScript("OnClick", function()
			if not (IsLeftShiftKeyDown() or IsRightShiftKeyDown()) then
				table.wipe(CheckRoles.SetDataList)
			end
			CheckRoles:SendCommMessage("CheckRolesRequest", "request", "GUILD", "ALERT")
		end)
		dataview.requestguild:SetScript("OnEvent", function(self)
			if IsInGuild() then
				self:Enable()
			else
				self:Disable()
			end
		end)
		dataview.requestguild:GetScript("OnEvent")(dataview.requestguild)
		dataview.requestguild:RegisterEvent("PLAYER_GUILD_UPDATE")
		dataview.requestguild:RegisterForClicks("LeftButtonUp")
		dataview.requestguild:SetPoint("TOPLEFT", dataview.requestraid, "BOTTOMLEFT", 0, -15)
		dataview.requestguild:SetPoint("TOPRIGHT", dataview.requestraid, "BOTTOMRIGHT", 0, -15)
		dataview.requestguild:SetHeight(20)
		function dataview:Update()
			for _, f in ipairs(self.players.entries) do f:Hide() end
			local cnt = 1
			for name, data in pairs(CheckRoles.SetDataList) do
				local b = self.players.entries[cnt]
				b:Show()
				b:SetData(name, data.c, data)
				cnt = cnt+1
			end
			self.players.scroll:SetHeight((cnt-1)*20)
			self.players.scrollframe:UpdateScrollChildRect()
			self.players.scrollframe:SetVerticalScroll(0)
		end
	end
	do
		self.GearView = CreateFrame("Frame", nil, UIParent)
		local gearview = self.GearView
		gearview:SetFrameStrata("HIGH")
		gearview:SetScript("OnHide", function(self) if self.ShouldShowDataViewOnClose then CheckRoles.DataView:Show() self.ShouldShowDataViewOnClose = false end if CheckRoles.DataView.specs.curHilight then CheckRoles.DataView.specs.curHilight:UnlockHighlight() CheckRoles.DataView.specs.curHilight = nil end end)
		gearview:Hide()
		gearview:SetPoint("CENTER")
		gearview.title = CreateTitleBar(gearview, "Gear viewer")
		gearview:SetWidth(300)
		gearview:SetHeight(400)
		gearview:SetBackdrop({
			bgFile="Interface\\Tooltips\\UI-Tooltip-Background",
			edgeFile="Interface\\Tooltips\\UI-Tooltip-Border",
			tile=true,
			tileSize=16,
			edgeSize=16,
			insets={top=4,bottom=4,left=4,right=4},
		})
		gearview:SetBackdropColor(0.5,0.5,0.5)
		gearview.dropdown = CreateDropdown(gearview)
		gearview.dropdown:SetPoint("TOPLEFT", gearview.title, "BOTTOMLEFT", 5, 0)
		gearview.dropdown:SetPoint("TOPRIGHT", gearview.title, "BOTTOMRIGHT", -38, 0)
		gearview.dropdown:SetHeight(30)
		gearview.deletesetdummy = CreateFrame("Button", nil, gearview)
		gearview.deletesetdummy:SetPoint("LEFT", gearview.dropdown, "RIGHT", 2, 0)
		gearview.deletesetdummy:SetSize(30,30)
		gearview.deletesetdummy:SetDisabledTexture("Interface\\BUTTONS\\UI-GroupLoot-Pass-Up")
		if not gearview.deletesetdummy:GetDisabledTexture():SetDesaturated(true) then
			gearview.deletesetdummy:GetDisabledTexture():SetVertexColor(0.5,0.5,0.5)
		end
		gearview.deletesetdummy:Disable()
		gearview.dressup = CreateFrame("DressUpModel", nil, gearview)
		gearview.dressup:SetPoint("TOP", gearview.dropdown, "BOTTOM", 16, 0)
		gearview.dressup:SetSize(223,297)
		local su = gearview.dressup.SetUnit
		local ru = gearview.dressup.RefreshUnit
		function gearview.dressup:SetUnit(unit)
			if UnitExists(unit) then
				self.RefreshUnit = ru
				su(self, unit)
			else
				self.RefreshUnit = function() end
				self:ClearModel()
			end
		end
		gearview.equipbuttons = {}
		function gearview:UpdateGear()
			self.dressup:RefreshUnit()
			self.dressup:Undress()
			for _, btn in pairs(self.equipbuttons) do
				if btn.link then
					self.dressup:TryOn(btn.link)
				end
			end
		end
		for slotID, textureData in LoopAllEquipSlots() do
			local equipbtn = CreateFrame("Button", nil, gearview)
			do
				local snt = equipbtn.SetNormalTexture
				function equipbtn:SetNormalTexture(...)
					self:SetDisabledTexture(...)
					snt(self, ...)
				end
			end
			equipbtn:SetSize(35,35)
			equipbtn:SetNormalTexture("Interface\\PaperDoll\\UI-PaperDoll-Slot-"..textureData)
			equipbtn:SetHighlightTexture("Interface\\Buttons\\ButtonHilight-Square")
			equipbtn.slot = slotID
			equipbtn:SetScript("OnEnter", function(self)
				local gt = GameTooltip
				gt:SetOwner(self, "ANCHOR_RIGHT")
				gt:ClearLines()
				if self.link then
					gt:SetHyperlink(self.link)
				else
					gt:AddLine(self.slot, 0.8,0.8,0.2)
				end
				gt:Show()
			end)
			equipbtn:SetScript("OnLeave", function()
				GameTooltip:Hide()
			end)
			gearview.equipbuttons[slotID] = equipbtn
		end
		local btns = gearview.equipbuttons
		btns.Head:SetPoint("TOPRIGHT", gearview.dressup, "TOPLEFT")
		btns.Neck:SetPoint("TOP", btns.Head, "BOTTOM")
		btns.Shoulder:SetPoint("TOP", btns.Neck, "BOTTOM")
		btns.Cloak:SetPoint("TOP", btns.Shoulder, "BOTTOM")
		btns.Chest:SetPoint("TOP", btns.Cloak, "BOTTOM")
		btns.Shirt:SetPoint("TOP", btns.Chest, "BOTTOM")
		btns.Tabard:SetPoint("TOP", btns.Shirt, "BOTTOM")
		btns.Bracers:SetPoint("TOP", btns.Tabard, "BOTTOM")
		btns["Main Hand"]:SetPoint("TOPLEFT", gearview.dressup, "BOTTOMLEFT")
		btns["Off-hand"]:SetPoint("TOP", gearview.dressup, "BOTTOM")
		btns.Relic:SetPoint("TOPRIGHT", gearview.dressup, "BOTTOMRIGHT")
		btns.Ranged:SetPoint("TOPRIGHT", gearview.dressup, "BOTTOMRIGHT")
		btns.Gloves:SetPoint("TOPLEFT", gearview.dressup, "TOPRIGHT")
		btns.Belt:SetPoint("TOP", btns.Gloves, "BOTTOM")
		btns.Legs:SetPoint("TOP", btns.Belt, "BOTTOM")
		btns.Boots:SetPoint("TOP", btns.Legs, "BOTTOM")
		btns["Top Finger"]:SetPoint("TOP", btns.Boots, "BOTTOM")
		btns["Bottom Finger"]:SetPoint("TOP", btns["Top Finger"], "BOTTOM")
		btns["Top trinket"]:SetPoint("TOP", btns["Bottom Finger"], "BOTTOM")
		btns["Bottom trinket"]:SetPoint("TOP", btns["Top trinket"], "BOTTOM")
		function gearview:ApplySet(data, specstring, hasRelic)
			for slot, texdata in LoopAllEquipSlots() do
				if data[slot] then
					btns[slot]:SetNormalTexture(data[slot].tex)
					btns[slot].link = data[slot].link
				else
					btns[slot]:SetNormalTexture("Interface\\PaperDoll\\UI-PaperDoll-Slot-"..texdata)
					btns[slot].link = nil
				end
			end
			self:UpdateGear()
			self.dropdown.text:SetText(specstring)
			if hasRelic then
				btns.Relic:Show()
				btns.Ranged:Hide()
			else
				btns.Ranged:Show()
				btns.Relic:Hide()
			end
		end
	end
	do
		self.ActionSelect = CreateFrame("Frame", nil, UIParent)
		local actionselect = self.ActionSelect
		actionselect:SetFrameStrata("DIALOG")
		if CheckRoles_Data.HasHadFirstLoad then
			actionselect:Hide()
		else
			CheckRoles_Data.HasHadFirstLoad = true
		end
		actionselect:SetPoint("CENTER")
		actionselect:SetBackdrop({
			bgFile="Interface\\DialogFrame\\UI-DialogBox-Background",
			edgeFile="Interface\\DialogFrame\\UI-DialogBox-Border",
			tileSize=32,
			tile=true,
			insets={
				top=12,
				bottom=12,
				right=12,
				left=12,
			},
		})
		actionselect:SetSize(300, 190)
		actionselect.header = actionselect:CreateTexture()
		--actionselect.header:SetPoint("CENTER", actionselect, "TOP")
		actionselect.header:SetPoint("TOP", 0, 12)
		actionselect.header:SetTexture("Interface\\DialogFrame\\UI-DialogBox-Header")
		actionselect.header:SetSize(200,68)
		actionselect.header.text = actionselect:CreateFontString()
		--actionselect.header.text:SetAllPoints(actionselect.header)
		actionselect.header.text:SetPoint("TOP", actionselect.header, "TOP", 0, -14)
		actionselect.header.text:SetFontObject(GameFontNormal)
		actionselect.header.text:SetText("CheckRoles")
		local line = actionselect:CreateFontString()
		line:SetFont("Fonts\\FRIZQT__.TTF", 10, "OUTLINE")
		line:SetJustifyH("LEFT")
		line:SetPoint("TOPLEFT", 15, -30)
		line:SetText("Welcome to CheckRoles!\nYou should do the following steps to get started:")
		actionselect.creategearset = CreateFrame("Button", nil, actionselect)
		for textype, state in LoopTextures() do
			actionselect.creategearset["Set"..textype.."Texture"](actionselect.creategearset, "Interface\\Buttons\\UI-Panel-Button-"..state)
			actionselect.creategearset["Get"..textype.."Texture"](actionselect.creategearset):SetTexCoord(0,0,0,0.6875,0.625,0,0.625,0.6875)
		end
		actionselect.creategearset:SetNormalFontObject(GameFontHighlight)
		actionselect.creategearset:SetText("Create one or more sets of gear")
		actionselect.creategearset:SetScript("OnClick", function()
			CheckRoles.GearEdit:Show()
			CheckRoles.GearEdit.ShouldShowActionSelectOnHide = true
			CheckRoles.ActionSelect:Hide()
		end)
		actionselect.creategearset:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
			GameTooltip:ClearLines()
			GameTooltip:AddLine("/roles gear", 0.8,0.8,0.2)
			GameTooltip:AddLine("/roles sets", 0.8,0.8,0.2)
			GameTooltip:AddLine("/roles equipment", 0.8,0.8,0.2)
			GameTooltip:AddLine("/roles equip", 0.8,0.8,0.2)
			GameTooltip:Show()
		end)
		actionselect.creategearset:SetScript("OnLeave", function()
			GameTooltip:Hide()
		end)
		actionselect.creategearset:SetPoint("TOPLEFT", line, "BOTTOMLEFT", 0, -5)
		actionselect.creategearset:SetSize(270, 30)
		actionselect.assignroles = CreateFrame("Button", nil, actionselect)
		for textype, state in LoopTextures() do
			actionselect.assignroles["Set"..textype.."Texture"](actionselect.assignroles, "Interface\\Buttons\\UI-Panel-Button-"..state)
			actionselect.assignroles["Get"..textype.."Texture"](actionselect.assignroles):SetTexCoord(0,0,0,0.6875,0.625,0,0.625,0.6875)
		end
		actionselect.assignroles:SetNormalFontObject(GameFontHighlight)
		actionselect.assignroles:SetText("Assign the sets you created to roles")
		actionselect.assignroles:SetScript("OnClick", function()
			CheckRoles.RoleSelect:Show()
			CheckRoles.RoleSelect.ShouldShowActionSelectOnHide = true
			CheckRoles.ActionSelect:Hide()
		end)
		actionselect.assignroles:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
			GameTooltip:ClearLines()
			GameTooltip:AddLine("/roles roleselect", 0.8,0.8,0.2)
			GameTooltip:AddLine("/roles select", 0.8,0.8,0.2)
			GameTooltip:AddLine("/roles role", 0.8,0.8,0.2)
			GameTooltip:AddLine("/roles roles", 0.8,0.8,0.2)
			GameTooltip:Show()
		end)
		actionselect.assignroles:SetScript("OnLeave", function()
			GameTooltip:Hide()
		end)
		actionselect.assignroles:SetPoint("TOPLEFT", actionselect.creategearset, "BOTTOMLEFT", 0, -5)
		actionselect.assignroles:SetSize(270, 30)
		actionselect.requestdata = CreateFrame("Button", nil, actionselect)
		for textype, state in LoopTextures() do
			actionselect.requestdata["Set"..textype.."Texture"](actionselect.requestdata, "Interface\\Buttons\\UI-Panel-Button-"..state)
			actionselect.requestdata["Get"..textype.."Texture"](actionselect.requestdata):SetTexCoord(0,0,0,0.6875,0.625,0,0.625,0.6875)
		end
		actionselect.requestdata:SetNormalFontObject(GameFontHighlight)
		actionselect.requestdata:SetText("Request data from your guild or raid")
		actionselect.requestdata:SetScript("OnClick", function()
			CheckRoles.DataView:Show()
			CheckRoles.DataView.ShouldShowActionSelectOnHide = true
			CheckRoles.ActionSelect:Hide()
		end)
		actionselect.requestdata:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
			GameTooltip:ClearLines()
			GameTooltip:AddLine("/roles data", 0.8,0.8,0.2)
			GameTooltip:AddLine("/roles request", 0.8,0.8,0.2)
			GameTooltip:Show()
		end)
		actionselect.requestdata:SetScript("OnLeave", function()
			GameTooltip:Hide()
		end)
		actionselect.requestdata:SetPoint("TOPLEFT", actionselect.assignroles, "BOTTOMLEFT", 0, -5)
		actionselect.requestdata:SetSize(270, 30)
		actionselect.close = CreateFrame("Button", nil, actionselect)
		for textype, state in LoopTextures() do
			actionselect.close["Set"..textype.."Texture"](actionselect.close, "Interface\\Buttons\\UI-Panel-Button-"..state)
			actionselect.close["Get"..textype.."Texture"](actionselect.close):SetTexCoord(0,0,0,0.6875,0.625,0,0.625,0.6875)
		end
		actionselect.close:SetNormalFontObject(GameFontHighlightSmall)
		actionselect.close:SetText("Close this window")
		actionselect.close:SetScript("OnClick", function()
			actionselect:Hide()
		end)
		actionselect.close:SetScript("OnEnter", function(self)
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
			GameTooltip:ClearLines()
			GameTooltip:AddLine("You can always bring this window up again!", 0.8,0.8,0.2)
			GameTooltip:AddLine("/checkroles",0.8,0.8,0.2)
			GameTooltip:AddLine("/roles",0.8,0.8,0.2)
			GameTooltip:Show()
		end)
		actionselect.close:SetScript("OnLeave", function()
			GameTooltip:Hide()
		end)
		actionselect.close:SetPoint("BOTTOMRIGHT", -10, 10)
		actionselect.close:SetSize(120, 20)
	end
	SlashCmdList["CHECKROLES"] = function(input)
		input = input:lower()
		if input == "roles" or input == "roleselect" or input == "select" or input == "role" or input == "roles" then
			CheckRoles.RoleSelect:Show()
			return
		end
		if input == "gear" or input == "sets" or input == "equipment" or input == "equip" then
			CheckRoles.GearEdit:Show()
			return
		end
		if input == "data" or input == "request" then
			CheckRoles.DataView:Show()
			return
		end
		CheckRoles.ActionSelect:Show()
	end
	SLASH_CHECKROLES1 = "/checkroles"
	SLASH_CHECKROLES2 = "/roles"
end

--------------------------------------------------
-- update all dropdowns that display a set list --
--------------------------------------------------
function CheckRoles:UpdateAllSetDropdowns()
	self.GearEdit:UpdateDropdown()
	for _, f in pairs(CheckRoles.RoleSelect.roleframes) do
		for _, f2 in pairs(f.specs) do
			f2:UpdateDropdown()
		end
	end
end

----------------------------------------
-- fetch spec locale info if possible --
----------------------------------------
CheckRoles.speclocal = {}
do
	local spec = 1
	local name, icon = GetTalentTabInfo(spec)
	while name do
		CheckRoles.speclocal[spec] = {name=name, icon=icon}
		spec=spec+1
		name, icon = GetTalentTabInfo(spec)
	end
	if spec == 1 then
		-- this is an initial login. register for the earliest point talent info is available
		CheckRoles:RegisterEvent("PLAYER_ALIVE")
	else
		-- this is an reload. register for addon_loaded so saved variables can be checked
		CheckRoles:RegisterEvent("ADDON_LOADED")
	end
end

-------------------------------------------------------------
-- main event handler for item scanning and initialization --
-------------------------------------------------------------
CheckRoles:SetScript("OnEvent", function(self,event,...)
	local args = {...}
	if event == "PLAYER_ALIVE" then
		local spec = 1
		local name, icon = GetTalentTabInfo(spec)
		while name do
			CheckRoles.speclocal[spec] = {name=name, icon=icon}
			spec=spec+1
			name,icon = GetTalentTabInfo(spec)
		end
		CheckRoles_Data = UpdateSavedVariable(CheckRoles_Data)
		self:ScanEquipment()
		self:ScanBagContents()
		self:InitGUI()
		self:UnregisterEvent("PLAYER_ALIVE")
	end
	if event == "ADDON_LOADED" and args[1] == "CheckRoles" then
		CheckRoles_Data = UpdateSavedVariable(CheckRoles_Data)
		self:ScanEquipment()
		self:ScanBagContents()
		self:InitGUI()
		self:UnregisterEvent("ADDON_LOADED")
	end
	if event == "BAG_UPDATE" then
		self:ScanBagContents()
	end
	if event == "BANKFRAME_CLOSED" then
		self.BankOpened = false
	end
	if event == "BANKFRAME_OPENED" then
		self.BankOpened = true
		self:ScanBankContents()
	end
	if event == "UNIT_INVENTORY_CHANGED" and UnitIsUnit("player", args[1]) then
		self:ScanEquipment()
	end
end)
CheckRoles:RegisterEvent("UNIT_INVENTORY_CHANGED")
CheckRoles:RegisterEvent("BAG_UPDATE")
CheckRoles:RegisterEvent("BANKFRAME_CLOSED")
CheckRoles:RegisterEvent("BANKFRAME_OPENED")

---------------------
-- embed libraries --
---------------------
LibStub("AceComm-3.0"):Embed(CheckRoles)
LibStub("AceSerializer-3.0"):Embed(CheckRoles)

---------------------------
-- incoming comm handler --
---------------------------
CheckRoles:RegisterComm("CheckRolesData", function(prefix, message, dist, sender)
	local wasSuccess, class, texture, role, spec, specname, setdata = CheckRoles:Deserialize(LibCompress:Decompress(LibCompress:GetAddonEncodeTable():Decode(message)))
	if not wasSuccess then
		print("|cffff0000-CheckRoles-|r ERROR trying to deserialize set data from "..sender.."!")
	end
	CheckRoles.SetDataList[sender] = CheckRoles.SetDataList[sender] or {c=class}
	CheckSet(CheckRoles.SetDataList[sender], role, spec, {texture=texture, name=specname, data=setdata})
	CheckRoles.DataView:Update()
end)
CheckRoles:RegisterComm("CheckRolesRequest", function(prefix, message, dist, sender)
	for role in LoopRoles() do
		if CheckRoles_Data.roles[role] then
			for spec=1, GetNumTalentTabs() do
				if CheckRoles_Data.roles[role][spec] and CheckRoles_Data.roles[role][spec].IsEnabled and CheckRoles_Data.roles[role][spec].setName then
					local msg = LibCompress:GetAddonEncodeTable():Encode(LibCompress:Compress(CheckRoles:Serialize((select(2,UnitClass("player"))),CheckRoles.speclocal[spec].icon,role,spec, CheckRoles.speclocal[spec].name,CheckRoles_Data.sets[CheckRoles_Data.roles[role][spec].setName])))
					CheckRoles:SendCommMessage("CheckRolesData", msg, "WHISPER", sender, "BULK")
				end
			end
		end
	end
end)

---------------------
-- init data table --
---------------------
CheckRoles.SetDataList = {}

------------------------
-- inventory scanning --
------------------------
function CheckRoles:ScanBagContents()
	if not CheckRoles_Data then return end
	-- wipe current data - we're always re-scanning everything
	table.wipe(CheckRoles_Data.items.bags)
	-- loop through bags...
	for bag=0, NUM_BAG_SLOTS do
		-- ... and slots
		for slot=1, GetContainerNumSlots(bag) do
			local itemID, itemLink = (GetContainerItemID(bag, slot)), (GetContainerItemLink(bag,slot))
			-- check if slot is not empty
			if itemID then
				-- get remainder of item data from id
				local _,_,quality,iLvL,reqLvL,_,_,_,equipslot,texture = GetItemInfo(itemID)
				--only save items that are equippable
				if equipslot then
					-- table is indexed by equipment "type" (INVTYPE_X)
					CheckRoles_Data.items.bags[equipslot] = CheckRoles_Data.items.bags[equipslot] or {}
					tinsert(CheckRoles_Data.items.bags[equipslot], {id=itemID,link=itemLink,qual=quality,ilvl=iLvL,reqlvl=reqLvL,tex=texture})
				end
			end
		end
	end
	-- scan bank if it is open too
	if self.BankOpened then
		self:ScanBankContents()
	end
end

-------------------
-- bank scanning --
-------------------
function CheckRoles:ScanBankContents()
	if not CheckRoles_Data then return end
	table.wipe(CheckRoles_Data.items.bank)
	-- just make sure that it really IS open
	if not self.BankOpened then return end
	-- loop through bags...
	for bag=NUM_BAG_SLOTS+1, NUM_BAG_SLOTS+NUM_BANKBAGSLOTS do
		-- ... and slots
		for slot=1, GetContainerNumSlots(bag) do
			local itemID, itemLink = (GetContainerItemID(bag, slot)), (GetContainerItemLink(bag,slot))
			-- check if slot is not empty
			if itemID then
				-- get remainder of item data from id
				local _,_,quality,iLvL,reqLvL,_,_,_,equipslot,texture = GetItemInfo(itemID)
				--discard non-equipable
				if equipslot then
					-- table is indexed by equipment "type" (INVTYPE_X)
					CheckRoles_Data.items.bank[equipslot] = CheckRoles_Data.items.bank[equipslot] or {}
					tinsert(CheckRoles_Data.items.bank[equipslot], {id=itemID,link=itemLink,qual=quality,ilvl=iLvL,reqlvl=reqLvL,tex=texture})
				end
			end
		end
	end
	--blizzard sucks at numbering
	for slot=1, GetContainerNumSlots(-1) do
		local itemID, itemLink = (GetContainerItemID(-1, slot)), (GetContainerItemLink(-1,slot))
		-- check if slot is not empty
		if itemID then
			-- get remainder of item data from id
			local _,_,quality,iLvL,reqLvL,_,_,_,equipslot,texture = GetItemInfo(itemID)
			--y'know, remove non-equipable
			if equipslot then
				-- table is indexed by equipment "type" (INVTYPE_X)
				CheckRoles_Data.items.bank[equipslot] = CheckRoles_Data.items.bank[equipslot] or {}
				tinsert(CheckRoles_Data.items.bank[equipslot], {id=itemID,link=itemLink,qual=quality,ilvl=iLvL,reqlvl=reqLvL,tex=texture})
			end
		end
	end
end

------------------------
-- equipment scanning --
------------------------
function CheckRoles:ScanEquipment()
	if not CheckRoles_Data then return end
	table.wipe(CheckRoles_Data.items.equipment)
	for invSlot=1,19 do
		local itemID, itemLink=(GetInventoryItemID("player", invSlot)), (GetInventoryItemLink("player", invSlot))
		-- check if slot is not empty
		if itemID then
			local _,_,quality,iLvL,reqLvL,_,_,_,equipslot,texture = GetItemInfo(itemID)
			CheckRoles_Data.items.equipment[equipslot]=CheckRoles_Data.items.equipment[equipslot] or {}
			tinsert(CheckRoles_Data.items.equipment[equipslot], {id=itemID, link=itemLink, qual=quality, ilvl=iLvL,reqlvl=reqLvL,tex=texture})
		end
	end
end



--[[
	wip data:
	get links for an blizzard equip set:
	for slot, data in ipairs(GetEquipmentSetLocations(setname)) do local eq, bn, bg, s, b = EquipmentManager_UnpackLocation(data) print(slot, bg and GetContainerItemLink(b,s) or GetInventoryItemLink("player",s)) end
--]]